import json
from typing import TypeVar, Callable, Any, TYPE_CHECKING

T = TypeVar("T")


class GenericConfigMixin:
    def __init__(self, value: T, *, typecast: Callable[[Any], Any] = json.loads):
        super().__init__(value)  # type: ignore
        self.__typecast__ = typecast

    def __instancecheck__(self, value) -> bool:
        return False


class ListConfig(GenericConfigMixin, list):
    if TYPE_CHECKING:  # pragma: no cover

        def __new__(cls, value: T, typecast: Callable[[Any], Any]) -> T:
            ...

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, list)


class DictConfig(GenericConfigMixin, dict):
    if TYPE_CHECKING:  # pragma: no cover

        def __new__(cls, value: T, typecast: Callable[[Any], Any]) -> T:
            ...

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, dict)



from json import JSONEncoder
from typing import Any, List, MutableMapping

from configalchemy.field import Field

ConfigType = MutableMapping[str, Any]


class ConfigMetaItem:
    def __init__(self, priority: int, value: Any):
        self.priority = priority
        self.value = value

    def __repr__(self) -> str:
        return f"ConfigMetaItem(priority={self.priority}, value={self.value})"

    def __str__(self) -> str:
        return repr(self)


class ConfigMeta:
    def __init__(self, default_value: Any, field: Field):
        self.field = field
        self.value_list: List[ConfigMetaItem] = [ConfigMetaItem(0, default_value)]

    @property
    def value(self) -> Any:
        return self.value_list[-1].value

    def set(self, priority: int, value: Any) -> None:
        value = self.field.validate(value)
        item = ConfigMetaItem(priority, value)
        length = len(self.value_list)
        for index in range(length - 1, 0, -1):
            if self.value_list[index - 1].priority <= priority:
                self.value_list.insert(index, item)
                break
        else:
            self.value_list.insert(1, item)

    def __repr__(self) -> str:
        return repr(self.value)

    def __str__(self) -> str:
        return str(self.value)


class ConfigMetaJSONEncoder(JSONEncoder):
    def default(self, o: ConfigMeta) -> Any:
        if isinstance(o, ConfigMeta):
            return o.value
        super().default(o)



from typing import Type, Any, Union, Optional

AnyType = Optional[Type[Any]]


class ValidateException(Exception):
    def __init__(self, name: str, value: Any):
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"""{self.name}'s value is invalid:

value: {self.value} 
type {type(self.value)}
"""

    def __str__(self) -> str:
        return repr(self)


class Field:
    __slots__ = ("name", "value_type", "default_value")

    def __init__(self, *, name: str, value_type: AnyType, default_value: Any):
        self.name = name
        self.value_type = value_type or type(default_value)
        self.default_value = default_value
        self.prepare()

    def prepare(self) -> None:
        origin = getattr(self.value_type, "__origin__", None)
        if origin is None:
            # field is not "typing" object eg. Union etc.
            return
        if origin is Union:
            self.value_type = self.value_type.__args__
            return

    def validate(self, value: Any) -> Any:
        if self._instance_check(value):
            return value
        else:
            try:
                return self._typecast(value)
            except Exception as e:
                raise ValidateException(self.name, value) from e

    def _instance_check(self, value: Any) -> bool:
        if getattr(self.default_value, "__instancecheck__", None):
            return self.default_value.__instancecheck__(value)
        return isinstance(value, self.value_type)

    def _typecast(self, value: Any) -> Any:
        if getattr(self.default_value, "__typecast__", None):
            return self.default_value.__typecast__(value)
        if isinstance(self.value_type, tuple):
            return self.value_type[0](value)
        else:
            return self.value_type(value)



import unittest
from typing import Optional
from unittest.mock import patch, Mock

from configalchemy.ext.generic_config import ListConfig, DictConfig
from configalchemy.field import Field, ValidateException


class FieldTestCase(unittest.TestCase):
    def test_validate(self):
        int_field = Field(name="TEST", default_value=0, value_type=int)
        for value in [b"0", "0"]:
            self.assertEqual(0, int_field.validate(value))
        with self.assertRaises(ValidateException) as e:
            int_field.validate(".0")
        self.assertIn(".0", str(e.exception))
        self.assertIn(str(type(".0")), str(e.exception))
        self.assertEqual("TEST", e.exception.name)
        self.assertEqual(".0", e.exception.value)

    def test_union_type(self):
        optional_field = Field(
            name="TEST", default_value=None, value_type=Optional[int]
        )
        with patch.object(Field, "_typecast") as typecast:
            self.assertEqual(1, optional_field.validate(1))
            self.assertFalse(typecast.called)
        self.assertEqual(1, optional_field.validate("1"))

    def test_generic_field(self):
        value = ["1", "2"]
        typecast = Mock(return_value=value)
        list_config = ListConfig(value, typecast=typecast)
        generic_field = Field(name="TEST", default_value=list_config, value_type=None)
        with patch.object(Field, "_typecast") as _typecast:
            self.assertEqual(value, generic_field.validate(value))
            self.assertFalse(_typecast.called)
        self.assertEqual(value, generic_field.validate("typecast"))
        typecast.assert_called_with("typecast")

        value = {"1": 1}
        typecast = Mock(return_value=value)
        dict_config = DictConfig(value, typecast=typecast)
        generic_field = Field(name="TEST", default_value=dict_config, value_type=None)
        with patch.object(Field, "_typecast") as _typecast:
            self.assertEqual(value, generic_field.validate(value))
            self.assertFalse(_typecast.called)
        self.assertEqual(value, generic_field.validate("typecast"))
        typecast.assert_called_with("typecast")


if __name__ == "__main__":
    unittest.main()



import unittest

from configalchemy import get_current_config, BaseConfig
from configalchemy import configalchemy


class GlobalAPITestCase(unittest.TestCase):
    def test_get_current_config(self):
        configalchemy._current_config = None

        class DefaultConfig(BaseConfig):
            TEST = "test"

        current_config = get_current_config(DefaultConfig)
        self.assertEqual("test", current_config.TEST)

        DefaultConfig().TEST = "inited"
        current_config = get_current_config(DefaultConfig)
        self.assertEqual("inited", current_config.TEST)


if __name__ == "__main__":
    unittest.main()
